from cffi import FFI
from os.path import join
from sysconfig import get_config_var

ffibuilder = FFI()

ffibuilder.cdef(r"""
    typedef struct LikNormMachine LikNormMachine;
    enum Lik {
        BERNOULLI,
        BINOMIAL,
        POISSON,
        EXPONENTIAL,
        GAMMA,
        GEOMETRIC
    };

    LikNormMachine* create_machine(int);
    void apply1d(LikNormMachine *, enum Lik, double *, double *, double *,
                 size_t, double *, double *, double *);
    void apply2d(LikNormMachine *, enum Lik, double *, double *, double *,
                 double *, size_t, double *, double *, double *);
    void destroy_machine(LikNormMachine *);
""")

ffibuilder.set_source("limix_inference.liknorm._ffi",
r"""
    #include "liknorm/liknorm.h"

    enum Lik {
        BERNOULLI,
        BINOMIAL,
        POISSON,
        EXPONENTIAL,
        GAMMA,
        GEOMETRIC
    };

    LikNormMachine* create_machine(int n) { return liknorm_create_machine(n); }
    void destroy_machine(LikNormMachine *machine) { liknorm_destroy_machine(machine); }

    typedef void lik1d(LikNormMachine*, double);
    typedef void lik2d(LikNormMachine*, double, double);

    void* set_lik[] = {liknorm_set_bernoulli,
                       liknorm_set_binomial,
                       liknorm_set_poisson,
                       liknorm_set_exponential,
                       liknorm_set_gamma,
                       liknorm_set_geometric};

    void apply1d(LikNormMachine *machine,
                 enum Lik lik, double *x, double *tau, double *eta,
                 size_t size, double *log_zeroth, double *mean,
                 double *variance)
    {
        size_t i;
        for (i = 0; i < size; ++i)
        {
            ((lik1d*) set_lik[lik])(machine, x[i]);
            liknorm_set_prior(machine, tau[i], eta[i]);
            liknorm_integrate(machine, log_zeroth+i, mean+i, variance+i);
        }
    }

    void apply2d(LikNormMachine *machine,
                 enum Lik lik, double *x0, double *x1, double *tau,
                 double *eta, size_t size, double *log_zeroth, double *mean,
                 double *variance)
    {
        size_t i;
        for (i = 0; i < size; ++i)
        {
            ((lik2d*) set_lik[lik])(machine, x0[i], x1[i]);
            liknorm_set_prior(machine, tau[i], eta[i]);
            liknorm_integrate(machine, log_zeroth+i, mean+i, variance+i);
        }
    }
""", libraries=['liknorm'],
     library_dirs=[join(get_config_var('prefix'), 'lib')],
     include_dirs=[join(get_config_var('prefix'), 'include')])

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



from __future__ import absolute_import, division, unicode_literals

import logging

from numpy import clip, full
from numpy.linalg import lstsq

from ...lmm import FastLMM
from ...liknorm import LikNormMachine
from ._ep import EP


class ExpFamEP(EP):
    r"""Expectation Propagation for exponential family distributions.

    Models

    .. math::

        y_i ~|~ z_i \sim \text{ExpFam}(y_i ~|~ \mu_i = g(z_i))

    for

    .. math::

        \mathbf z \sim \mathcal N\big(~~ \mathrm M^\intercal \boldsymbol\beta;~
            \sigma_b^2 \mathrm Q_0 \mathrm S_0 \mathrm Q_0^{\intercal} +
                    \sigma_{\epsilon}^2 \mathrm I ~~\big)

    where :math:`\mathrm Q_0 \mathrm S_0 \mathrm Q_0^\intercal`
    is the economic eigen decomposition of a semi-definite positive matrix,
    :math:`g(\cdot)` is a link function, and :math:`\text{ExpFam}(\cdot)` is
    an exponential-family distribution.

    For convenience, let us define the following variables:

    .. math::
        :nowrap:

        \begin{eqnarray}
            \sigma_b^2          & = & v (1-\delta) \\
            \sigma_{\epsilon}^2 & = & v \delta\\
            \mathbf m           & = & \mathrm M \boldsymbol\beta \\
            \mathrm K           & = & \sigma_b^2 \mathrm Q_0 \mathrm S_0
                                      \mathrm Q_0^{\intercal} +
                                      \sigma_{\epsilon}^2 \mathrm I
        \end{eqnarray}

    Args:
        prodlik (object): likelihood product.
        covariates (array_like): fixed-effect covariates :math:`\mathrm M`.
        Q0 (array_like): eigenvectors of positive eigenvalues.
        Q1 (array_like): eigenvectors of zero eigenvalues.
        S0 (array_like): positive eigenvalues.
    """
    def __init__(self, prodlik, covariates, Q0, Q1, S0):
        super(ExpFamEP, self).__init__(covariates, Q0, S0, True)
        self._logger = logging.getLogger(__name__)

        self._Q1 = Q1
        self._machine = LikNormMachine(500)
        self._likname = prodlik.name

        h2, m = _initialize(prodlik, covariates, Q0, Q1, S0)

        n = prodlik.sample_size

        self._phenotype = prodlik
        self._tbeta = lstsq(self._tM, full(n, m))[0]
        self.delta = 1 - h2
        self.v = 1.

    def _tilted_params(self):
        ctau = self._cav_tau
        ceta = self._cav_eta
        lmom0 = self._loghz
        self._machine.moments(self._likname, self._phenotype.ytuple, ceta, ctau,
                              lmom0, self._hmu, self._hvar)

    @property
    def genetic_variance(self):
        r"""Genetic variance.

        Returns:
            :math:`\sigma_b^2`.
        """
        return self.sigma2_b

    @property
    def environmental_variance(self):
        r"""Environmental variance.

        Returns:
            :math:`\sigma_{\epsilon}^2`.
        """
        return self.sigma2_epsilon

    @property
    def heritability(self):
        r"""Narrow-sense heritability.

        Returns:
            :math:`\sigma_b^2/(\sigma_a^2+\sigma_b^2+\sigma_{\epsilon}^2)`.
        """
        total = self.genetic_variance + self.covariates_variance
        total += self.environmental_variance
        return self.genetic_variance / total


def _initialize(prodlik, covariates, Q0, Q1, S0):
    y = prodlik.to_normal()
    flmm = FastLMM(y, Q0, Q1, S0, covariates=covariates)
    flmm.learn()
    gv = flmm.genetic_variance
    nv = flmm.environmental_variance
    h2 = gv / (gv + nv)
    return clip(h2, 0.01, 0.9), flmm.m



# from __future__ import division
#
# from numpy import array, dot, ones, pi, sqrt, zeros
# from numpy.random import RandomState
# from numpy.testing import assert_almost_equal
#
# from numpy_sugar.linalg import economic_qs_linear
#
# from limix_inference._ep import ExpFamEP
# from limix_inference.lik import BinomialProdLik
# from limix_inference.link import LogitLink
#
#
# def test_binomial_lml():
#     n = 3
#     M = ones((n, 1)) * 1.
#     G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
#     (Q, S0) = economic_qs_linear(G)
#     nsuccesses = array([1., 0., 1.])
#     ntrials = array([1., 1., 1.])
#     lik = BinomialProdLik(ntrials, LogitLink())
#     lik.nsuccesses = nsuccesses
#     ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1)
#     ep.beta = array([1.])
#     assert_almost_equal(ep.beta, array([1.]))
#     ep.v = 1.
#     ep.delta = 0
#     assert_almost_equal(ep.lml(), -2.3202659215368935)
#
#
# def test_binomial_gradient_over_v():
#     n = 3
#     M = ones((n, 1)) * 1.
#     G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
#     (Q, S0) = economic_qs_linear(G)
#     nsuccesses = array([1., 0., 1.])
#     ntrials = array([1., 1., 1.])
#     lik = BinomialProdLik(ntrials, LogitLink())
#     lik.nsuccesses = nsuccesses
#     ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1)
#     ep.beta = array([1.])
#     assert_almost_equal(ep.beta, array([1.]))
#     ep.v = 1.
#     ep.delta = 0.
#
#     analytical_gradient = ep._gradient_over_v()
#
#     lml0 = ep.lml()
#     step = 1e-5
#     ep.v = ep.v + step
#     lml1 = ep.lml()
#
#     empirical_gradient = (lml1 - lml0) / step
#
#     assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)
#
#     ep.v = 0.5
#     ep.delta = 0.0
#
#     analytical_gradient = ep._gradient_over_v()
#
#     lml0 = ep.lml()
#     step = 1e-5
#     ep.v = ep.v + step
#     lml1 = ep.lml()
#
#     empirical_gradient = (lml1 - lml0) / step
#
#     assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)
#
#
# def test_binomial_gradient_over_delta():
#     n = 3
#     M = ones((n, 1)) * 1.
#     G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
#     (Q, S0) = economic_qs_linear(G)
#     nsuccesses = array([1., 0., 1.])
#     ntrials = array([1., 1., 1.])
#     lik = BinomialProdLik(ntrials, LogitLink())
#     lik.nsuccesses = nsuccesses
#     ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1.0)
#     ep.beta = array([1.])
#     assert_almost_equal(ep.beta, array([1.]))
#     ep.v = 1.
#     ep.delta = 0.5
#
#     analytical_gradient = ep._gradient_over_delta()
#
#     lml0 = ep.lml()
#     step = 1e-5
#     ep.delta = ep.delta + step
#     lml1 = ep.lml()
#
#     empirical_gradient = (lml1 - lml0) / step
#
#     assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)
#
#
# def test_binomial_gradient_over_both():
#     n = 3
#     M = ones((n, 1)) * 1.
#     G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
#     (Q, S0) = economic_qs_linear(G)
#     nsuccesses = array([1., 0., 1.])
#     ntrials = array([1., 1., 1.])
#     lik = BinomialProdLik(ntrials, LogitLink())
#     lik.nsuccesses = nsuccesses
#     ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1.0)
#     ep.beta = array([1.])
#     assert_almost_equal(ep.beta, array([1.]))
#     ep.v = 1.5
#     ep.delta = 0.3
#
#     analytical_gradient = ep._gradient_over_both()[0]
#
#     lml0 = ep.lml()
#     step = 1e-5
#     ep.v = ep.v + step
#     lml1 = ep.lml()
#
#     empirical_gradient = (lml1 - lml0) / step
#
#     assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)
#
#     ep.v = 1.5
#     ep.delta = 0.3
#
#     analytical_gradient = ep._gradient_over_both()[1]
#
#     lml0 = ep.lml()
#     step = 1e-5
#     ep.delta = ep.delta + step
#     lml1 = ep.lml()
#
#     empirical_gradient = (lml1 - lml0) / step
#
#     assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)
#
#
# def test_binomial_optimize():
#     random = RandomState(139)
#     nsamples = 30
#     nfeatures = 31
#
#     G = random.randn(nsamples, nfeatures) / sqrt(nfeatures)
#
#     u = random.randn(nfeatures)
#
#     z = 0.1 + 2 * dot(G, u) + random.randn(nsamples)
#
#     ntrials = random.randint(10, 500, size=nsamples)
#
#     y = zeros(nsamples)
#     for i in range(len(ntrials)):
#         y[i] = sum(z[i] + random.logistic(
#             scale=pi / sqrt(3), size=ntrials[i]) > 0)
#     (Q, S0) = economic_qs_linear(G)
#
#     M = ones((nsamples, 1))
#     lik = BinomialProdLik(ntrials, LogitLink())
#     lik.nsuccesses = y
#     ep = ExpFamEP(lik, M, Q[0], Q[1], S0)
#     ep.optimize()
#
#     assert_almost_equal(ep.lml(), -144.23818408071736, decimal=3)
#     assert_almost_equal(ep.sigma2_b, 1.13419722224, decimal=2)
#     assert_almost_equal(ep.sigma2_epsilon, 0.122855059169, decimal=2)
#     assert_almost_equal(ep.beta[0], 0.110667132173, decimal=2)
#     assert_almost_equal(ep.heritability, 0.902267343224, decimal=2)
#
#
# if __name__ == '__main__':
#     __import__('pytest').main([__file__, '-s'])



from __future__ import division

from numpy import exp
from numpy import clip
from numpy import atleast_2d

from numpy_sugar import is_all_finite

from optimix import maximize_scalar
from optimix import Function
from optimix import Scalar

from ._core import FastLMMCore


class FastLMM(Function):
    def __init__(self, y, Q0, Q1, S0, covariates=None):
        super(FastLMM, self).__init__(logistic=Scalar(0.0))

        if not is_all_finite(y):
            raise ValueError("There are non-finite values in the phenotype.")

        self._flmmc = FastLMMCore(y, covariates, Q0, Q1, S0)
        self.set_nodata()

    @property
    def M(self):
        return self._flmmc.M

    @M.setter
    def M(self, v):
        self._flmmc.M = v

    def copy(self):
        o = FastLMM.__new__(FastLMM)
        super(FastLMM, o).__init__(logistic=Scalar(self.get('logistic')))
        o._flmmc = self._flmmc.copy()
        o.set_nodata()
        return o

    def _delta(self):
        v = clip(self.get('logistic'), -20, 20)
        x = 1 / (1 + exp(-v))
        return clip(x, 1e-5, 1 - 1e-5)

    @property
    def heritability(self):
        t = (self.fixed_effects_variance + self.genetic_variance +
             self.environmental_variance)
        return self.genetic_variance / t

    @property
    def fixed_effects_variance(self):
        return self._flmmc.m.var()

    @property
    def genetic_variance(self):
        return self._flmmc.scale * (1 - self._flmmc.delta)

    @property
    def environmental_variance(self):
        return self._flmmc.scale * self._flmmc.delta

    @property
    def beta(self):
        return self._flmmc.beta

    @property
    def m(self):
        return self._flmmc.m

    def learn(self, progress=None):
        maximize_scalar(self, progress=progress)
        self._flmmc.delta = self._delta()

    def value(self):
        self._flmmc.delta = self._delta()
        return self._flmmc.lml()

    def lml(self):
        self._flmmc.delta = self._delta()
        return self._flmmc.lml()

    def predict(self, X, covariates, Xp, trans=None):
        covariates = atleast_2d(covariates)
        Xp = atleast_2d(Xp)
        if trans is not None:
            Xp = trans.transform(Xp)
        Cp = Xp.dot(X.T)
        Cpp = Xp.dot(Xp.T)
        return self._flmmc.predict(covariates, Cp, Cpp)



from numpy import eye
from collections import Sequence


class Fruits(object):
    __slots__ = ['shape']

    def __init__(self, shape=None):
        if shape is None or isinstance(shape, Sequence):
            self.shape = shape
        else:
            self.shape = (shape, )

    def __eq__(self, that):
        assert self.ndim == that.ndim
        c = 1.0 * (isinstance(self, type(that)))
        if self.ndim == 0:
            return c
        elif self.ndim == 1:
            return c * eye(self.shape[0], that.shape[0])
        assert False

    @property
    def ndim(self):
        if self.shape is None:
            return 0
        return len(self.shape)


class Apples(Fruits):
    def __init__(self, shape=None):
        super(Apples, self).__init__(shape)


class Oranges(Fruits):
    def __init__(self, shape=None):
        super(Oranges, self).__init__(shape)



"""
*******************
Linear Mixed Models
*******************

- :class:`.FastLMM`
- :class:`.SlowLMM`

Linear mean
^^^^^^^^^^^

.. autoclass:: FastLMM
  :members:

Offset mean
^^^^^^^^^^^

.. autoclass:: SlowLMM
  :members:
"""

from ._fastlmm import FastLMM
from ._slowlmm import SlowLMM



# from __future__ import division
#
# from numpy import array, dot, empty, ones, sqrt
# from numpy.random import RandomState
# from numpy.testing import assert_almost_equal, assert_allclose
#
# from numpy_sugar.linalg import economic_qs_linear
#
# from limix_inference._ep import ExpFamEP
# from limix_inference.lik import BernoulliProdLik
# from limix_inference.link import LogitLink
#
#
# def test_bernoulli_lml():
#     n = 3
#     M = ones((n, 1)) * 1.
#     G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
#     (Q, S0) = economic_qs_linear(G)
#     y = array([1., 0., 1.])
#     lik = BernoulliProdLik(LogitLink())
#     lik.outcome = y
#     ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1.0)
#     ep.beta = array([1.])
#     assert_almost_equal(ep.beta, array([1.]))
#     ep.v = 1.
#     ep.delta = 0.
#     assert_almost_equal(ep.lml(), -2.3202659215368935)
#     assert_almost_equal(ep.sigma2_epsilon, 0)
#     assert_almost_equal(ep.sigma2_b, 1)
#
#
# def test_bernoulli_gradient_over_v():
#     n = 3
#     M = ones((n, 1)) * 1.
#     G = array([[1.2, 3.4], [-.1, 1.2], [0.0, .2]])
#     (Q, S0) = economic_qs_linear(G)
#     y = array([1., 0., 1.])
#     lik = BernoulliProdLik(LogitLink())
#     lik.outcome = y
#     ep = ExpFamEP(lik, M, Q[0], Q[1], S0 + 1.0)
#     ep.beta = array([1.])
#     assert_almost_equal(ep.beta, array([1.]))
#     ep.v = 1.
#     ep.delta = 0.
#
#     analytical_gradient = ep._gradient_over_v()
#
#     lml0 = ep.lml()
#     step = 1e-5
#     ep.v = ep.v + step
#     lml1 = ep.lml()
#
#     empirical_gradient = (lml1 - lml0) / step
#
#     assert_almost_equal(empirical_gradient, analytical_gradient, decimal=4)
#
#
# def test_bernoulli_optimize():
#     random = RandomState(139)
#     nsamples = 100
#     nfeatures = nsamples + 10
#
#     G = random.randn(nsamples, nfeatures) / sqrt(nfeatures)
#
#     M = ones((nsamples, 1))
#
#     u = random.randn(nfeatures)
#
#     z = 0.1 + dot(G, u) + 0.5 * random.randn(nsamples)
#
#     y = empty(nsamples)
#     y[z > 0] = 1
#     y[z <= 0] = 0
#
#     (Q, S0) = economic_qs_linear(G)
#
#     lik = BernoulliProdLik(LogitLink())
#     lik.outcome = y
#     ep = ExpFamEP(lik, M, Q[0], Q[1], S0)
#     ep.optimize()
#     assert_allclose(ep.lml(), -67.68161933698035, rtol=1e-5)
#     assert_allclose(ep.heritability, 0.931995296712, rtol=1e-5)
#     assert_allclose(ep.beta[0], -0.16082171171359558, rtol=1e-5)
#
#
# if __name__ == '__main__':
#     __import__('pytest').main([__file__, '-s'])



from numpy import ones
from numpy.random import RandomState
from numpy.testing import assert_almost_equal

from optimix import check_grad

from limix_inference.mean import OffsetMean


def test_value():
    random = RandomState(0)
    mean = OffsetMean()
    offset = random.randn()
    mean.offset = offset

    n = 10
    oarr = offset * ones(n)

    assert_almost_equal(mean.value(n), oarr)


def test_gradient():
    mean = OffsetMean()

    n = 10

    def func(x):
        mean.offset = x[0]
        return mean.value(n)

    def grad(x):
        mean.offset = x[0]
        return [mean.derivative_offset(n)]

    assert_almost_equal(check_grad(func, grad, [2.0]), 0, decimal=6)


if __name__ == '__main__':
    __import__('pytest').main([__file__, '-s'])
